!Brute-force numerical integration in FORTRAN using trapezoid method

MODULE numint
IMPLICIT NONE



CONTAINS

    FUNCTION trapezoid(h, a, b)
    IMPLICIT NONE
    
    DOUBLE PRECISION                             :: h, a, b !h: height, a&b: function values at x=a and x=b
    DOUBLE PRECISION                             :: trapezoid
    
    trapezoid=0.5*h*(a+b)
    
    END FUNCTION trapezoid
    
    
    
    SUBROUTINE trapint(funame, lowerbound, upperbound, tolerance, result)
    IMPLICIT NONE
    
    DOUBLE PRECISION, INTENT(in)                 :: lowerbound, upperbound
    DOUBLE PRECISION                             :: tolerance, h, u, l, approxres, stopcrit, work, sigma, argsig, prevres
    DOUBLE PRECISION                             :: funame
    DOUBLE PRECISION, INTENT(out)                :: result
    INTEGER                          :: workspace, nstep, subdiv, nn, jj
    EXTERNAL funame
    
    workspace=1000 !maximum number of steps
    
    !***INITIALIZING PROCEDURE*****
    nstep=-1 !number of steps in numerical integration
    
    u=upperbound
    l=lowerbound
    
    h=(u-l)
    
    subdiv=1
    work=0.5*(funame(l)+funame(u))
	approxres=work
    
    DO nn=0, workspace
    
       nstep=nstep+1
       
       h=0.5*h
       
       sigma=0
       
       DO jj=1, subdiv
       
          argsig=l+(2*jj-1)*h
          sigma=sigma+funame(argsig)
          
       ENDDO
       
       work=work+sigma
       prevres=approxres
       approxres=h*work
       subdiv=2*subdiv
       
       stopcrit=ABS(approxres-prevres)
       
       IF (stopcrit .le. (ABS(approxres)*tolerance+tolerance)) THEN
       
           OPEN(UNIT=99, FILE="numintlog.txt", STATUS="unknown")
           WRITE(99,*) 'Error is: ', stopcrit
           CLOSE(99)
           
           EXIT
           
       ENDIF
       
    ENDDO
    
    IF (nstep .gt. workspace) THEN
    
       OPEN(UNIT=9, FILE="ERRORLOG.txt", STATUS="unknown")
       WRITE(9,*) 'ERROR: Integral was not computed within specified precision'
       CLOSE(9)
       
    ELSE
    
       result=approxres
       
    ENDIF
    
    END SUBROUTINE trapint
    
    
END MODULE numint
    
