!File with the full GvM2 simulation algorithm

PROGRAM sim_gvm2

USE constants!list of modules used
USE numint

IMPLICIT NONE

	!list of variables


	DOUBLE PRECISION				:: res, tol, up, low, f, norm, theta, fval, x1, x2, y1, y2, xsect, platzhalter
	DOUBLE PRECISION				:: gvm2, dummyx, dummyf, dgvm2, ddgvm2, dummyapp, ntot, twopi=2.0d0*pi, dk1, dk2, dm1, dm2
	DOUBLE PRECISION, DIMENSION(:), ALLOCATABLE	:: rc, ic, rz, iz, wrk, rctp, rcex, ictp, icex, wrktp, wrkex, rzex, rztp, izex, iztp
	DOUBLE PRECISION, DIMENSION(:), ALLOCATABLE	:: truetp, trueex, falsetp, falseex
	DOUBLE PRECISION, DIMENSION(2)			:: m, h
	DOUBLE PRECISION, DIMENSION(10)			:: approxpts, fapprox
	INTEGER						:: jj, deg, ier, nzero, npoints
	INTEGER						:: nchain, nacc, nsim
	CHARACTER					:: yn
	LOGICAL						:: MCMCyn, ROUyn, RADIAN, ARyn, IARyn
	CHARACTER(LEN=64)				:: string
	external f, gvm2, dgvm2, ddgvm2




	!main body of the program

	OPEN(20, FILE='logfile.log', STATUS='unknown')

	!input

	WRITE(*,*) 'Please input the parameters for the program'
	WRITE(*,*) 'Is the input for mu1 and mu2 in radian (y/n)?'
	READ(*,*) yn

	IF (yn .eq. 'y') THEN

		RADIAN = .TRUE.

	ELSE

		RADIAN = .FALSE.

	ENDIF

	yn='n'


	!*********************** Reading the parameters of the GvM2 *******************
 
	WRITE(*,*) 'Please input kappa1:'
	READ(*,*) dk1

	IF (dk1 .le. 0d0) THEN

		WRITE(*,*) 'Parameter kappa1 is out of bound (must be >0)'

		STOP

	ELSE

		k1=dk1
		WRITE(*,*) 'kappa1 set at: kappa1=', k1

	ENDIF

	WRITE(*,*) 'Please input kappa2:'
	READ(*,*) dk2

	IF (dk2 .le. 0d0) THEN

		WRITE(*,*) 'Parameter kappa2 is out of bound (must be >0)'

		STOP

	ELSE

		k2=dk2
		WRITE(*,*) 'kappa2 set at: kappa2=', k2

	ENDIF

	WRITE(*,*) 'Please input mu1:'
	READ(*,*) dm1

	IF (RADIAN) THEN

		IF (dm1 .lt. 0d0 .or. dm1 .ge. 2*pi) THEN

			WRITE(*,*) 'Parameter mu1 is out of bound (must be in [0,2Pi)'

			STOP

		ELSE

			m1=dm1
			WRITE(*,*) 'mu1 set at: mu1=', m1

		ENDIF

	ELSE

		IF (dm1 .lt. 0d0 .or. dm1 .ge. 360.0) THEN

			WRITE(*,*) 'Parameter mu1 is out of bound (must be in [0,2Pi)'

			STOP

		ELSE

			m1=dm1
			WRITE(*,*) 'mu1 set at: mu1=', m1

		ENDIF
 
	ENDIF  

	WRITE(*,*) 'Please input mu2:'
	READ(*,*) dm2

	IF (RADIAN) THEN

		IF (dm2 .lt. 0d0 .or. dm2 .ge. pi) THEN

			WRITE(*,*) 'Parameter mu2 is out of bound (must be in [0,Pi)'

			STOP

		ELSE

			m2=dm2
			WRITE(*,*) 'mu2 set at: mu2=', m2

		ENDIF

	ELSE

		IF (dm2 .lt. 0d0 .or. dm2 .ge. 180.0) THEN

			WRITE(*,*) 'Parameter mu2 is out of bound (must be in [0,Pi)'

			STOP

		ELSE

			m2=dm2
			WRITE(*,*) 'mu2 set at: mu2=', m2

		ENDIF

	ENDIF

	IF (.NOT. RADIAN) m1=m1*pi/180.0d0
	IF (.NOT. RADIAN) m2=m2*pi/180.0d0 

	WRITE(*,*) 'Parameters are: kappa1=', k1, ', kappa2=', k2, ', mu1=', m1, ', mu2=', m2, ' and delta=', MOD(m1-m2,pi)

	!**********************END READING PARAMETERS********************************




	!***************** Evaluating the extrema and turning points of the GvM2 ********************

	ALLOCATE(rctp(9))
	ALLOCATE(ictp(9))
	ALLOCATE(rcex(5))
	ALLOCATE(icex(5))

	ALLOCATE(rztp(8))
	ALLOCATE(iztp(8))
	ALLOCATE(rzex(4))
	ALLOCATE(izex(4))

	ALLOCATE(wrktp(37))
	ALLOCATE(wrkex(21))

	CALL firstderivzero(rcex, icex)
	WRITE(20,*) 'The coefficients for the first derivative polynomial are:'
	WRITE(20,*) rcex

	CALL secondderivzero(rctp, ictp)
	WRITE(20,*) 'The coefficients for the second derivative polynomial are:'
	WRITE(20,*) rctp

	!STOP ! Fehlermeldung aufgrund falscher Reihenfolge für die Koeffizienten => c0->c9, c1->c8 etc
	!WRITE(20,*) 'Flag'

	CALL polynomial_zero(4, rcex, nzero, rzex)
	CALL read_npoints(4, rzex, npoints)
	WRITE(20,*) 'Number of zeros:', npoints
	WRITE(20,*) 'Raw maxima:'
	DO jj=1,4

		WRITE(20,*) rzex

	END DO

	ALLOCATE(trueex(npoints))
	ALLOCATE(falseex(4-npoints))

	CALL true_zeros(4, rzex, npoints, trueex, falseex)
	CALL sort(trueex)
	WRITE(20,*) 'Extrema:'
	DO jj=1, npoints

		platzhalter=gvm2(trueex(jj))
		WRITE(20,*) trueex(jj), platzhalter

	END DO

	!STOP

	CALL polynomial_zero(8, rctp, nzero, rztp)
	CALL read_npoints(8, rztp, npoints)
	WRITE(20,*) 'Number of turning points:', npoints
	ALLOCATE(truetp(npoints))
	ALLOCATE(falsetp(8-npoints))

	CALL true_zeros(8, rztp, npoints, truetp, falsetp)
	CALL sort(truetp)
	WRITE(20,*) 'Turning points:'
	DO jj=1, npoints

		platzhalter=gvm2(truetp(jj))
		WRITE(20,*) truetp(jj), platzhalter
	
	END DO

	!Creating vector with points for the approximation:

	CALL critical_points(trueex, 4, truetp, 4, approxpts, fapprox)

	WRITE(20,*) 'Envelope support points:'
	DO jj=1, 10

		WRITE(*,*) jj, approxpts(jj), fapprox(jj)
		WRITE(20,*) jj, approxpts(jj), fapprox(jj)

	END DO

	OPEN(UNIT=23, FILE='gvmdat.txt', STATUS='unknown')

	DO jj=1, 5000

		dummyx=2*pi/(4999*1.0d0)*(jj-1)
		dummyf=gvm2(dummyx)
		dummyapp=gvm2approx(dummyx)

		WRITE(23,*) dummyx, dummyf, dummyapp

	END DO

	WRITE(*,*) 'Simulating from random_approxgvm2'
	DO jj=1, 1000

		!WRITE(*,*) jj
		CALL random_approxgvm2(dummyx,ntot)
		WRITE(20,*) dummyx

	END DO

	DEALLOCATE(rctp)
	DEALLOCATE(ictp)
	DEALLOCATE(rcex)
	DEALLOCATE(icex)
	DEALLOCATE(rztp)
	DEALLOCATE(iztp)
	DEALLOCATE(rzex)
	DEALLOCATE(izex)
	DEALLOCATE(wrktp)
	DEALLOCATE(wrkex)


	!simulated draws from GvM2 distribution:

	WRITE(*,*) 'Performing improved acceptance-rejection simulation'
	WRITE(*,*) 'Please input the number of desired accepted simulations:'
	READ(*,*) nacc
	CALL iacc_rej(nacc,gvm2)
	CALL trapint(gvm2,0.0d0,twopi,1d-6,norm)
	WRITE(*,*) 'The total surface area of the approximation is', ntot
	WRITE(*,*) 'The total surface area of the generalized von Mises is', norm
	WRITE(*,*) 'The efficiency is (GvM2/approx)', norm/ntot





	CONTAINS	!subroutines necessary for the program



	!subroutines computing the coefficients of the first and second derivative polynomials

	SUBROUTINE firstderivzero(rcoeff,icoeff)
	USE constants
	IMPLICIT NONE
	
	! Polynomial coefficients for the fourth-order polynomial needed to find the maxima and minima
	! from Gatto in Stat.Comput.
	
	DOUBLE PRECISION							:: delta, rho, diff
	DOUBLE PRECISION							:: c0, c1, c2, c3, c4
	DOUBLE PRECISION, DIMENSION(5), INTENT(OUT)				:: rcoeff, icoeff
	INTEGER									:: jj
	
	!OPEN(UNIT=1000, FILE="GvMparam.txt", STATUS="OLD")
	!READ(1000,*) k1, k2, m1, m2, delta
	!CLOSE(1000)
	
	rho=k1/(4.0*k2)
	delta=MOD(m1-m2,pi)
	diff=ABS(delta)
	
	IF (ABS(diff-0) .le. 1.0e-9 .or. ABS(diff-pi) .le. 1.0e-9 .or. ABS(diff-pi/2.0) .le. 1.0e-9) THEN
	
		c0=0.0
		c1=0.0
		c2=rho**2 - 1.0
		c3=0.0
		c4=1.0

	ELSE
	
		c0=(SIN(delta)*COS(delta))**2
		c1=2.0*rho*SIN(delta)*COS(delta)
		c2=rho**2 - 1.0
		c3=-4.0*rho*SIN(delta)*COS(delta)
		c4=1.0
		
	END IF
	
	OPEN(UNIT=2000, FILE="extremcoeff.txt", STATUS="UNKNOWN")
	WRITE(2000,*) c0, c1, c2, c3, c4
	
	rcoeff(1)=c4!c0
	rcoeff(2)=c3!c1
	rcoeff(3)=c2!c2
	rcoeff(4)=c1!c3
	rcoeff(5)=c0!c4
	icoeff(1)=0.0
	icoeff(2)=0.0
	icoeff(3)=0.0
	icoeff(4)=0.0
	icoeff(5)=0.0
	
	DO jj=1,5
	
		WRITE(2000,*) rcoeff(jj), icoeff(jj)
		
	END DO
	
	CLOSE(2000) 
	
	END SUBROUTINE firstderivzero
	
	SUBROUTINE secondderivzero(rcoeff,icoeff)
	USE constants
	IMPLICIT NONE
	
	! Polynomial coefficients for the eighth-order polynomial needed to determine the turning points
	! from my notes (thank MAPLE!)
	
	DOUBLE PRECISION							:: delta, rho, diff, diff0, diff2Pi, diffPihalf
	DOUBLE PRECISION							:: c0, c1, c2, c3, c4, c5, c6, c7, c8, cd, sd
	DOUBLE PRECISION, DIMENSION(9), INTENT(OUT) 				:: rcoeff, icoeff
	INTEGER									:: jj
	
	!OPEN(UNIT=1000, FILE="GvMparam.txt", STATUS="OLD")
	!READ(1000,*) k1, k2, m1, m2, delta
	!CLOSE(1000)
	
	rho=k1/(4.0*k2)

	delta=MOD(m1-m2,pi)
	
	diff=ABS(delta)
	diff0=ABS(delta-0)
	diff2Pi=ABS(delta-2*pi)
	diffPihalf=ABS(delta-pi/2.0)
	
	cd=COS(delta)
	sd=SIN(delta)
	
	IF (diffPihalf .le. 1.0e-9) THEN
	
		c0=-256.0*k2**3*sd**6 - 64.0*k2**2*sd**2 - k1**2 + 16.0*k2**2 -512.0*k2**4*sd**6 &
		&+ 256.0*k2**4*sd**8 +64*k2**2*sd**4+256.0*k2**4*sd**4 -128.0*k2**3*sd**2 +384.0*k2**3*sd**4
		c1=0.0
		c2=8.0*k1**2*k2 + k1**2 -64.0*k2**2 -128.0*k2**3 + 512.0*k2**3*sd**6 -96.0*k1**2*k2**2*sd**4 -16.0*k1**2*k2*sd**2 &
		&-768.0*k2**3*sd**4 -512.0*k2**4*sd**2 +512.0*k2**4*sd**4 + 512.0*k2**3*sd**2 +96.0*k1**2*k2**2*sd**2
		c3=0.0
		c4=-32.0*k1**2*k2**2 + 64.0*k2**2 + 384.0*k2**3 + 256.0*k2**4 + k1**4 + 256.0*k1**2*k2**2*sd**4 + 512.0*k2**4*sd**2 &
		&-512.0*k2**4*sd**4 -768.0*k2**3*sd**2 - 256.0*k1**2*k2**2*sd**2
		c5=0.0
		c6=32.0*k2**2*k1**2 - 256.0*k2**3 - 512.0*k2**4 - 256.0*k2**2*k1**2*sd**4 + 512.0*k2**3*sd**2 &
		&+ 256.0*k1**2*k2**2*sd**2
		c7=0.0
		c8=256.0*k2**4
		
	ELSEIF (diff0 .le. 1.0e-9 .or. diff2Pi .le. 1.0e-9) THEN
	
		c0=- k1**2 + 16.0*k2**2
		c1=0.0
		c2=8.0*k1**2*k2 + k1**2 -64.0*k2**2 -128.0*k2**3
		c3=0.0
		c4=-32.0*k1**2*k2**2 + 64.0*k2**2 + 384.0*k2**3 + 256.0*k2**4 + k1**4
		c5=0.0
		c6=32.0*k2**2*k1**2 - 256.0*k2**3 - 512.0*k2**4
		c7=0.0
		c8=256.0*k2**4	
	
	ELSE
	
		c0=-256.0*k2**3*sd**6 - 64.0*k2**2*sd**2 - k1**2 + 16.0*k2**2 -512.0*k2**4*sd**6 &
		&+ 256.0*k2**4*sd**8 +64*k2**2*sd**4+256.0*k2**4*sd**4 -128.0*k2**3*sd**2 +384.0*k2**3*sd**4
		
		c1=256.0*k1*k2**3*sd**3*cd + 32.0*k1*k2*sd*cd - 256.0*k1*k2**3*cd*sd**5
		
		c2=8.0*k1**2*k2 + k1**2 -64.0*k2**2 -128.0*k2**3 + 512.0*k2**3*sd**6 -96.0*k1**2*k2**2*sd**4 -16.0*k1**2*k2*sd**2 &
		&-768.0*k2**3*sd**4 -512.0*k2**4*sd**2 +512.0*k2**4*sd**4 + 512.0*k2**3*sd**2 +96.0*k1**2*k2**2*sd**2
		
		c3=-192.0*k1*k2**2*sd*cd -256.0*k1*k2**3*cd*sd + 384.0*k1*k2**2*sd**3*cd + 512.0*k1*k2**3*sd**5*cd &
		&+ 16.0*k1**3*k2*sd*cd - 32.0*k1*k2*sd*cd -512.0*k1*k2**3*sd**3*cd
		
		c4=-32.0*k1**2*k2**2 + 64.0*k2**2 + 384.0*k2**3 + 256.0*k2**4 + k1**4 + 256.0*k1**2*k2**2*sd**4 &
		&+ 512.0*k2**4*sd**2 -512.0*k2**4*sd**4 -768.0*k2**3*sd**2 - 256.0*k1**2*k2**2*sd**2
		
		c5=128.0*k1*k2**2*sd*cd - 32.0*k1**3*k2*sd*cd -256.0*k1*k2**2*sd**3*cd + 768.0*k1*k2**3*sd*cd
		
		c6=32.0*k2**2*k1**2 - 256.0*k2**3 - 512.0*k2**4 - 256.0*k2**2*k1**2*sd**4 + 512.0*k2**3*sd**2 &
		&+ 256.0*k1**2*k2**2*sd**2
		
		c7=-512.0*k1*k2**3*sd*cd
		
		c8=256.0*k2**4
		
	END IF
	
	OPEN(UNIT=3000, FILE="tpcoeff.txt", STATUS="UNKNOWN")
	WRITE(3000,*) c0, c1, c2, c3, c4, c5, c6, c7, c8
	
	rcoeff(1)=c8!c0
	rcoeff(2)=c7!c1
	rcoeff(3)=c6!c2
	rcoeff(4)=c5!c3
	rcoeff(5)=c4!c4
	rcoeff(6)=c3!c5
	rcoeff(7)=c2!c6
	rcoeff(8)=c1!c7
	rcoeff(9)=c0!c8
	icoeff(1)=0.0
	icoeff(2)=0.0
	icoeff(3)=0.0
	icoeff(4)=0.0
	icoeff(5)=0.0
	icoeff(6)=0.0
	icoeff(7)=0.0
	icoeff(8)=0.0
	
	DO jj=1,9
	
		WRITE(3000,*) rcoeff(jj), icoeff(jj)
		
	END DO
	
	CLOSE(3000) 

	END SUBROUTINE secondderivzero



	!subroutines computing the zeros of a polynomial


	SUBROUTINE polynomial_zero(deg, coeff, nzero, zeros)	!returns the DOUBLE PRECISION zeros of an n-th degree polynomial
	IMPLICIT NONE
	
	INTEGER, INTENT(in)				  			:: deg
	DOUBLE PRECISION, DIMENSION(deg+1)					:: coeff, tracf ! coefficients and - if necessary - transformed coefficients
	DOUBLE PRECISION, DIMENSION(deg)					:: rzero, izero 
	DOUBLE PRECISION, DIMENSION(deg+1)					:: rcoeff, icoeff
	DOUBLE PRECISION, DIMENSION(deg)					:: zeros
	DOUBLE PRECISION, DIMENSION(4*deg+5)			 		:: work
	DOUBLE PRECISION							:: placeholder
	INTEGER							  		:: nzero, jj, kk, ier, degwork, nzero_found
	LOGICAL							  		:: FAIL=.FALSE.
	
	!Statistics for ERROR analysis:
	OPEN(UNIT=999, FILE="polyzerolog.txt", STATUS="unknown")
	WRITE(999,*) 'Order input', deg
	WRITE(999,*) 'Coefficients:'
	
	  DO kk=0,deg
	  
		 WRITE(999,*) 'c(',kk,')', coeff(kk+1)
		 
	  END DO
	
	placeholder=coeff(1)
	!WRITE(*,*) 'Flag poly 1', placeholder
	IF (placeholder .eq. 1.0) THEN
	
		DO jj=1,deg+1
		
		  tracf(jj)=coeff(jj)
		  
		END DO
	
	ELSE
	
		DO jj=1,deg+1
		
		  tracf(jj)=coeff(jj)/placeholder
		  
		END DO
	
	END IF
	!WRITE(*,*) 'Flag poly 2', tracf(1)
	DO jj=1, deg+1
	
		rcoeff(jj)=DBLE(tracf(jj))
		icoeff(jj)=DBLE(0.0d0)					!allow only DOUBLE PRECISION constants
		
	END DO
	
	!degwork=deg
	
	CALL weierstrass(deg, rcoeff, rzero, izero)

	DO jj=1,deg!nzero_found!
	
		WRITE(999,*) rzero(jj), izero(jj)
		
	END DO
	!WRITE(*,*) 'Flag poly 4'
	nzero=0
	
	DO jj=1,deg
	!WRITE(*,*) 'Flag poly 4.1'
		IF (ABS(izero(jj)) .le. 1.0e-6) THEN
		!WRITE(*,*) 'Flag poly 4.2'
		  zeros(jj)=rzero(jj)
		  nzero=nzero+1
		!WRITE(*,*) 'Flag poly 4.3'	
		ELSE
		
		  zeros(jj)=3.0*pi
		  nzero=nzero
		  
		END IF
		
	END DO
	!WRITE(*,*) 'Flag poly 5'
	WRITE(999,*) 'There are', nzero, 'real zeros'
	WRITE(999,*) 'The real zeros are:'
	
	DO jj=1,deg

		WRITE(999,*) zeros(jj)
		
	END DO
	
	CLOSE(999)
	
	END SUBROUTINE polynomial_zero
	

	SUBROUTINE weierstrass(deg, rcoeff, zeror, zeroi)
	IMPLICIT NONE
	
	INTEGER									:: jj, kk, deg, nzero, mod_deg1, mod_deg2, mod_deg3, mod_deg4, mod_deg5	
	COMPLEX, DIMENSION(deg)							:: xpt
	DOUBLE PRECISION, DIMENSION(deg)					:: zeror, zeroi
	DOUBLE PRECISION							:: dummy1, dummy2, zerocond, inv_zerocond
	DOUBLE PRECISION, DIMENSION(deg+1)					:: rcoeff, inv_rcoeff 
	!coefficients in descending / ascending order
	COMPLEX, DIMENSION(:), ALLOCATABLE					:: modxpt
	DOUBLE PRECISION, DIMENSION(:), ALLOCATABLE				:: modcoeff, rzero, izero
	LOGICAL									:: flag1, flag2, flag3, flag4, flag5
	!flag1: zero is solution, flag2: substitution to quadratic equation, flag3: substitution to quartic equation
	!flag4: flag1 and flag2, flag5: flag1 and flag3
	LOGICAL, DIMENSION(deg+1)				:: zero_coeff, inv_zero_coeff
	
	!I/O
	
	OPEN(UNIT=1313, FILE="weierstrasslog.txt", STATUS="unknown")
	WRITE(1313,*) 'I/O ok'
	!Preparations
	
	mod_deg1=0
	mod_deg2=0
	mod_deg3=0
	mod_deg4=0
	mod_deg5=0
	nzero=0
	flag1=.FALSE.
	flag2=.FALSE.
	flag3=.FALSE.
	flag4=.FALSE.
	flag5=.FALSE.
	
	DO jj=1, deg+1
	
		zero_coeff(jj)=.FALSE.
		inv_zero_coeff(jj)=.FALSE.
		
	END DO
	
	DO jj=0, deg
	
		inv_rcoeff(jj+1)=rcoeff(deg+1-jj)
		WRITE(1313,*) 'Polynomial coefficients:'
		WRITE(1313,*) rcoeff(jj+1), inv_rcoeff(jj+1)
	
	END DO	
	
	DO jj=1,deg+1
	
		zerocond=ABS(rcoeff(jj)-0.0)
		inv_zerocond=ABS(inv_rcoeff(jj)-0.0)
		IF (zerocond .lt. 1.0e-9) zero_coeff(jj)=.TRUE.
		IF (inv_zerocond .lt. 1.0e-9) inv_zero_coeff(jj)=.TRUE.
		WRITE(1313,*) 'Coefficient (descending)', jj, 'is zero:', zero_coeff(jj)
		WRITE(1313,*) 'Coefficient (ascending)', jj, 'is zero:', inv_zero_coeff(jj)
		
	END DO
	WRITE(1313,*) 'Initialization ok'
	!Seed starting points in the complex plane
	
	DO jj=1, deg
	
		CALL random_number(dummy1)
		CALL random_number(dummy2)
		xpt(jj)=CMPLX(dummy1*jj, dummy2*jj)
		WRITE(1313,*) 'Starting points in the complex plane:'
		WRITE(1313,*) jj, xpt(jj)
		
	END DO
	WRITE(1313,*) 'Starting points ok'

	WRITE(1313,*) 'In iteration'
		
	ALLOCATE(rzero(deg))
	ALLOCATE(izero(deg))
		
	CALL iteration(deg, rcoeff, xpt)
	WRITE(1313,*) 'Successfully called iteration'
	DO jj=1,deg
		
		rzero(jj)=DBLE(xpt(jj))
		izero(jj)=AIMAG(xpt(jj))
		  
	END DO
	WRITE(1313,*) 'Iteration ok'

	DO jj=1,deg
		
		zeror(jj)=rzero(jj)
		zeroi(jj)=izero(jj)
		  
	END DO

	WRITE(1313,*) 'The zeros are:'
	DO jj=1,deg
	
		WRITE(1313,*) zeror(jj), zeroi(jj)
		
	END DO
		  
	CLOSE(1313)

	END SUBROUTINE weierstrass
	
	SUBROUTINE iteration(deg, coeff, zeros)
	IMPLICIT NONE
	
	INTEGER								:: deg, jj, kk, ll
	DOUBLE PRECISION, DIMENSION(deg+1)				:: coeff
	COMPLEX, DIMENSION(deg)						:: zeros, work
	DOUBLE PRECISION, DIMENSION(deg)				:: crit
	DOUBLE PRECISION						:: tol
	COMPLEX								:: numerator, polyval, corr
	
	
	OPEN(UNIT=888, FILE='iterationlog.txt', STATUS='unknown')
	tol=1.0d-9
	WRITE(888,*) 'Tolerance:', tol
	WRITE(888,*) 'Coefficients in descending order:', coeff
	
	IF (deg .EQ. 1) THEN
	
		zeros(1)=-coeff(2)/coeff(1)
	
	ELSE IF (deg .EQ. 2) THEN
	
		zeros(1)=(-coeff(2)+SQRT(coeff(2)**2-4.0*coeff(1)*coeff(3)))/(2.0*coeff(1))
		zeros(2)=(-coeff(2)-SQRT(coeff(2)**2-4.0*coeff(1)*coeff(3)))/(2.0*coeff(1))
	
	ELSE IF (deg .EQ. 3) THEN

	CALL cubic_solver(coeff, zeros)

	ELSE IF (deg .EQ. 4) THEN

	CALL quartic_solver(coeff,zeros)

	ELSE IF (deg .GE. 5) THEN

		DO ll=1,5000
		WRITE(888,*) ll, '. iteration step'
		  DO jj=1, deg
			  
			  work(jj)=zeros(jj)
			  numerator=1.0
			  
			  DO kk=1, deg

				 IF (kk .EQ. jj) THEN
					
					numerator=1.0*numerator
					
				 ELSE
				 
					numerator=numerator*(zeros(jj)-zeros(kk))
					WRITE(888,*) jj, kk
					WRITE(888,*) zeros(jj), zeros(kk)
				 END IF
				 WRITE(888,*) 'Numerator:', numerator
				 
			  END DO
			  polyval=poly(zeros(jj),coeff,deg)
			  WRITE(888,*) 'Polynomial value:', polyval
			  IF (ABS(polyval) .LT. 1.0d-9 .AND. ABS(numerator) .LT. 1.0d-9) THEN
			  
				 corr=0.0
				 
			  ELSE
			  
				 corr=polyval/numerator
				 
			  END IF
			  WRITE(888,*) 'Correction:', corr
			  zeros(jj)=zeros(jj)-corr
			  crit(jj)=ABS((zeros(jj)-work(jj))/work(jj))
			  WRITE(888,*) work(jj), zeros(jj)
			  WRITE(888,*) 'Criterion:', crit(jj)
			END DO
			
			IF (MAXVAL(crit) .LT. tol) EXIT !(1),crit(deg)
			IF (MAXVAL(crit) .GT. tol .AND. ll .EQ. 5000) THEN !(1),crit(deg)
			
			  WRITE(*,*) 'No root found'
			  
			END IF
			
		END DO  
		
	END IF

	END SUBROUTINE iteration

	
	FUNCTION poly(x,coeff,deg)
	COMPLEX							:: x, poly, wk
	INTEGER							:: jj, ex, deg
	DOUBLE PRECISION, DIMENSION(deg+1)			:: coeff
		
	wk=0
	WRITE(888,*) 'Polynomial determining function received:'
	WRITE(888,*) 'Degree=', deg
	DO jj=0, deg
		
		WRITE(888,*) jj+1, '. coefficient=', coeff(jj+1)
			
	END DO
	DO jj=1, deg+1
		
		ex=deg-jj+1
		wk=wk+coeff(jj)*x**ex
		WRITE(888,*) ex, wk
	END DO
		
	poly=wk
	WRITE(888,*) 'Polynomial yields'
	WRITE(888,*) x, poly
	END FUNCTION poly
		

	
	SUBROUTINE quartic_solver(coeff, zeros)
	IMPLICIT NONE
	DOUBLE PRECISION, DIMENSION(5), INTENT(IN)					:: coeff
	COMPLEX, DIMENSION(4), INTENT(OUT)						:: zeros
	DOUBLE PRECISION								:: a, b, c, p, q, dummy
	COMPLEX										:: u, r, y, w, posroot, negroot

	a=-3*0.125*coeff(2)**2/coeff(1)**2+coeff(3)/coeff(1)
	b=0.125*coeff(2)**3/coeff(1)**3-coeff(2)*coeff(3)/coeff(1)*0.5+coeff(4)/coeff(1)
	c=-3*3.90625d-3*coeff(2)**4/coeff(1)**4+0.0625*coeff(3)*coeff(2)**2/coeff(1)**3-coeff(4)*coeff(2)/coeff(1)**2*.25+coeff(5)/coeff(1)
	!WRITE(*,*) 'a=', a, 'b=', b, 'c=', c
	IF (ABS(b) .LE. 1.0d-12) THEN

		zeros(1)=-coeff(2)/coeff(1)*.25+SQRT(-a*.5+.5*SQRT(a**2-4*c))
		zeros(2)=-coeff(2)/coeff(1)*.25+SQRT(-a*.5-.5*SQRT(a**2-4*c))
		zeros(3)=-coeff(2)/coeff(1)*.25-SQRT(-a*.5+.5*SQRT(a**2-4*c))
		zeros(4)=-coeff(2)/coeff(1)*.25-SQRT(-a*.5-.5*SQRT(a**2-4*c))

	ELSE 

		p=-a**2/12.0-c
		q=-a**3/108.0+a*c/3.0-b**2*0.125
		r=-q*.5+SQRT(CMPLX(.25*q**2+p**3/27.0,0.0d0))
		u=r**(1.0/3.0)
		!WRITE(*,*) 'p=', p, 'q=', q, 'r=', r, 'u=', u
		IF (ABS(u) .LE. 1.0d-12) THEN

			y=-5*a/6.0-q**(1.0/3.0)

		ELSE		

			y=-5*a/6.0+u-p/u/3.0

		END IF
		w=SQRT(a+2*y)
		!WRITE(*,*) 'y=',y, 'w=', w

		posroot=SQRT(-(3*a+2*y+2*b/w))
		negroot=SQRT(-(3*a+2*y-2*b/w))
		!WRITE(*,*) 'positive root=', posroot, 'negative root=', negroot
		zeros(1)=-coeff(2)/coeff(1)*.25+0.5*(w+posroot)
		zeros(2)=-coeff(2)/coeff(1)*.25+0.5*(w-posroot)
		zeros(3)=-coeff(2)/coeff(1)*.25+0.5*(-w+negroot)
		zeros(4)=-coeff(2)/coeff(1)*.25+0.5*(-w-negroot)

	END IF

	END SUBROUTINE quartic_solver


	SUBROUTINE cubic_solver(coeff, zeros)
	IMPLICIT NONE
	DOUBLE PRECISION, DIMENSION(4), INTENT(IN)					:: coeff
	COMPLEX, DIMENSION(3), INTENT(OUT)						:: zeros
	COMPLEX										:: A, B, w1, w2, R
	DOUBLE PRECISION								:: Y, Z
	DOUBLE PRECISION, PARAMETER							:: sq3=SQRT(3.0), third=1.0/3.0

	w1=CMPLX(1.0d0,sq3)
	w2=CMPLX(1.0d0,-sq3)

	Y=2*coeff(2)**3-9*coeff(1)*coeff(2)*coeff(3)+27*coeff(1)**2*coeff(4)
	Z=coeff(2)**2-3*coeff(1)*coeff(3)

	R=SQRT(Y**2-4*Z**3)
	A=(0.5*(Y+R))**third
	B=(0.5*(Y-R))**third

	zeros(1)=-coeff(2)/coeff(1)*third-third/coeff(1)*(A+B)
	zeros(2)=-coeff(2)/coeff(1)*third+w1*third*.5/coeff(1)*A+w2*0.5*third/coeff(1)*B
	zeros(3)=-coeff(2)/coeff(1)*third+w2*third*.5/coeff(1)*A+w1*0.5*third/coeff(1)*B

	END SUBROUTINE cubic_solver
	


	!subroutines computing the true zeros from the complex zeros


	SUBROUTINE read_npoints(deg, zeros, npoints)
	IMPLICIT NONE
	
	INTEGER									:: deg, npoints, jj
	DOUBLE PRECISION, DIMENSION(deg)					:: zeros

	npoints=0
	
	DO jj=1,deg
	
		IF (zeros(jj) .le. 1.0 .and. zeros(jj) .ge. -1.0) THEN
	
		  npoints=npoints+1
!			
!			IF (deg .eq. 4) THEN
!			
!			  npoints=npoints+1
!			  
!			END IF
!			
		ELSE
		
		  npoints=npoints
		  
		END IF
		
	END DO

	END SUBROUTINE read_npoints
	
	
	SUBROUTINE true_zeros(deg, zeros, npoints, points, discarded)
	USE constants
	IMPLICIT NONE
	INTEGER									:: deg, npoints, jj, nwork,  ndisc, ll
	DOUBLE PRECISION, DIMENSION(deg)					:: zeros
	DOUBLE PRECISION, DIMENSION(npoints)					:: points
	DOUBLE PRECISION, DIMENSION(deg-npoints)				:: discarded
	DOUBLE PRECISION							:: delta, gt2pi, deriv_val, dderiv_val, dgvm2, ddgvm2
	DOUBLE PRECISION							:: eps, dplus, dminus
	EXTERNAL dgvm2, ddgvm2
	
	!OPEN(UNIT=1000, FILE="GvMparam.txt", STATUS="OLD")
	!READ(1000,*) k1, k2, m1, m2, delta
	!CLOSE(1000)
	
	nwork=0
	ndisc=0
	eps=pi/180.0d0
	
	IF (npoints .EQ. deg) THEN
	
		DO jj=1,deg
		
			IF (zeros(jj) .le. 1.0 .and. zeros(jj) .ge. -1.0) THEN
			  
			  nwork=nwork+1
			  
			  DO ll=1,deg
				 WRITE(20,*) 'Multiple zero?', zeros(jj), zeros(ll)
				 IF (ABS(zeros(jj)-zeros(ll)) .LT. 9.0e-5 .AND. ll .LT. jj) THEN
				 
					points(nwork)=pi-points(ll)+2*m1!pi-ASIN(zeros(jj))+m1!+MODULO(m1, 2*pi)
				 WRITE(20,*) 'yes'
				 WRITE(20,*) points(nwork)
				 EXIT
				 ELSE
				 WRITE(20,*) 'no'
					points(nwork)=ASIN(zeros(jj))+m1!+MODULO(m1, 2*pi)
					deriv_val=dgvm2(points(nwork))
					dderiv_val=ddgvm2(points(nwork))
				 WRITE(20,*) points(nwork), deriv_val, dderiv_val	
					IF (deg .EQ. 4) THEN
					
				dplus=dgvm2(points(nwork)+eps)
				dminus=dgvm2(points(nwork)-eps)
						!IF (ABS(deriv_val) .GT. 9.0e-5) THEN
						
						!	points(nwork)=pi-ASIN(zeros(jj))+MODULO(m1, 2*pi)
						  
						!END IF

				IF (SIGNUM(dplus) .EQ. SIGNUM(dminus)) points(nwork)=pi-points(nwork)+2*m1
						
					ELSE IF (deg .EQ. 8) THEN

				dplus=ddgvm2(points(nwork)+eps)
				dminus=ddgvm2(points(nwork)-eps)
					
						!IF (ABS(dderiv_val) .GT. 9.0e-5) THEN
						
						!	points(nwork)=pi-ASIN(zeros(jj))+MODULO(m1,2*pi)
						  
						!END IF

				IF (SIGNUM(dplus) .EQ. SIGNUM(dminus)) points(nwork)=pi-points(nwork)+2*m1
						
					END IF
					
				 WRITE(20,*) points(nwork)  
				 END IF
				 
			  END DO

! 			  nwork=nwork+1
! 			  points(nwork)=MODULO(pi-ASIN(zeros(jj))+m1, pi)
						
	!			IF (deg .eq. 4) THEN
	!			
	!			  nwork=nwork+1
	!			  points(nwork)=pi-ASIN(zeros(jj))+m1!MODULO(pi-ASIN(zeros(jj))+m1, 2.0*pi)!
	!			  
	!			END IF
			  
			ELSE
			
			  ndisc=ndisc+1
			  discarded(ndisc)=zeros(jj)
			  
			END IF
			
		END DO
		
	ELSE
	
		WRITE(20,*) 'Raw TP:', zeros
		DO jj=1,deg
		
			IF (zeros(jj) .le. 1.0 .and. zeros(jj) .ge. -1.0) THEN
			  
			  WRITE(20,*) 'Entry'
			  nwork=nwork+1
			  points(nwork)=ASIN(zeros(jj))+m1!+MODULO(m1, 2*pi)
				  deriv_val=dgvm2(points(nwork))
			  dderiv_val=ddgvm2(points(nwork))
			  WRITE(20,*) points(nwork), deriv_val, dderiv_val
			  IF (deg .EQ. 4) THEN
					
			dplus=dgvm2(points(nwork)+eps)
			dminus=dgvm2(points(nwork)-eps)

				 !IF (ABS(deriv_val) .GT. 9.0e-5) THEN
						
				 !	points(nwork)=pi-ASIN(zeros(jj))+MODULO(m1, 2*pi)
						  
				 !END IF

			IF (SIGNUM(dplus) .EQ. SIGNUM(dminus)) points(nwork)=pi-points(nwork)+2*m1
			  
			  ELSE IF (deg .EQ. 8) THEN
					
			dplus=ddgvm2(points(nwork)+eps)
			dminus=ddgvm2(points(nwork)-eps)

				 !IF (ABS(dderiv_val) .GT. 9.0e-5) THEN
						
				 !	points(nwork)=pi-ASIN(zeros(jj))+m1!+MODULO(m1,2*pi)
				 !WRITE(20,*) 'Switched value'
				 !END IF
			IF (SIGNUM(dplus) .EQ. SIGNUM(dminus)) points(nwork)=pi-points(nwork)+2*m1				  
	 
			  END IF
			  WRITE(20,*) 'Exit', points(nwork)
! 			  nwork=nwork+1
! 			  points(nwork)=pi-ASIN(zeros(jj))+MODULO(m1, 2*pi)
! 						
	!			IF (deg .eq. 4) THEN
	!			
	!			  nwork=nwork+1
	!			  points(nwork)=pi-ASIN(zeros(jj))+m1!MODULO(pi-ASIN(zeros(jj))+m1, 2.0*pi)!
	!			  
	!			END IF
			  
			ELSE
			
			  ndisc=ndisc+1
			  discarded(ndisc)=zeros(jj)
			  
			END IF
			
		END DO
		
	END IF

	DO jj=1, npoints
		
	  gt2pi=ABS(points(jj)-2*pi)
	  WRITE(20,*) points(jj), gt2pi
	  IF (points(jj) .GE. 2*pi) THEN
	  
		 points(jj)=points(jj)-2*pi
	  
	  ELSE IF (gt2pi .LT. 1.0d-9) THEN
	  
		 points(jj)=0.0
		
	  END IF
	  
	  IF (ABS(points(jj)) .LT. 1.0d-9) THEN
	  
		 points(jj)=0.0
		 
	  END IF

	  IF (points(jj) .LT. 0.0d0) THEN

		points(jj)=2*pi+points(jj)

	  END IF
	  
	END DO
		
	WRITE(20,*) 'zeros:', points


	END SUBROUTINE true_zeros
	

	INTEGER FUNCTION SIGNUM(x)
	DOUBLE PRECISION		  :: x

	IF (x .GT. 0.0d0) THEN

		SIGNUM=1

	ELSE IF (x .LT. 0.0d0) THEN

		SIGNUM=-1

	ELSE

		SIGNUM=0

	END IF

	END FUNCTION
		



	!subroutine computing the critical points of the GvM2


	SUBROUTINE critical_points(ex, nex, tp, ntp, critp, fcritp)
	IMPLICIT NONE
	INTEGER									:: nex, ntp, ncritp, jj, locmax(1), locmax1,locmax2, kk
	DOUBLE PRECISION, DIMENSION(nex)					:: ex
	DOUBLE PRECISION, DIMENSION(ntp)					:: tp
	DOUBLE PRECISION, DIMENSION(nex+ntp)					:: work_critp, f_work_critp
	DOUBLE PRECISION, DIMENSION(nex+ntp+2)					:: critp, fcritp
	DOUBLE PRECISION							:: gvm2, dgvm2, ddgvm2
	DOUBLE PRECISION							:: tan_max, h_max, tan_tp, h_tp, wrk, fwrk
	DOUBLE PRECISION							:: tpmax1sect, max1tpsect, tpmax2sect, max2tpsect
	EXTERNAL gvm2, dgvm2, ddgvm2
	
	DO jj=1, nex
	
		work_critp(jj)=ex(jj)
		WRITE(20,*) jj, work_critp(jj), ex(jj)
	END DO
	
	DO jj=nex+1, nex+ntp
	
		work_critp(jj)=tp(jj-nex)
		WRITE(20,*) jj, work_critp(jj), tp(jj-nex)
	END DO
	
	CALL sort(work_critp)
	
	DO jj=1, nex+ntp
	
		f_work_critp(jj)=gvm2(work_critp(jj))
		!WRITE(*,*) jj, work_critp(jj), f_work_critp(jj)
	END DO
	
	locmax=MAXLOC(f_work_critp)
	locmax1=locmax(1)
	
	IF (locmax1 .LT. 5) THEN
	
		locmax2=locmax1+4
		
	ELSE
	
		locmax2=locmax1-4
		
	END IF
	
	IF (locmax1 .GT. locmax2) THEN
	
		locmax1=locmax2
		locmax2=locmax(1)
		
	END IF
	
	WRITE(20,*) 'Maxima located at:', locmax1, locmax2
	!For the first maximum:
	WRITE(20,*) 'For the first maximum:'

	IF (locmax1 .EQ. 1) THEN
	
		tan_tp=dgvm2(work_critp(8)-2.0*pi)
		h_tp=gvm2(work_critp(8)-2.0*pi)-tan_tp*(work_critp(8)-2.0*pi)
		WRITE(20,*) 'TP:', tan_tp, h_tp, work_critp(8)-2*pi
	ELSE
	
		tan_tp=dgvm2(work_critp(locmax1-1))
		h_tp=gvm2(work_critp(locmax1-1))-tan_tp*work_critp(locmax1-1)
		WRITE(20,*) 'TP:', tan_tp, h_tp
	END IF
	
	tan_max=0.0
	h_max=gvm2(work_critp(locmax1))
	WRITE(20,*) tan_tp, h_tp, tan_max, h_max
	CALL intersect(tan_tp, h_tp, tan_max, h_max, tpmax1sect)
	WRITE(20,*) tan_tp, h_tp, tan_max, h_max, tpmax1sect
	IF (tpmax1sect .LT. 0.0) THEN
	
		tpmax1sect=tpmax1sect+2*pi
		
	END IF
	tan_tp=dgvm2(work_critp(locmax1+1))
	h_tp=gvm2(work_critp(locmax1+1))-tan_tp*work_critp(locmax1+1)
	WRITE(20,*) tan_tp, h_tp, tan_max, h_max
	CALL intersect(tan_max, h_max, tan_tp, h_tp, max1tpsect)
	WRITE(20,*) tan_tp, h_tp, tan_max, h_max, max1tpsect
	IF (max1tpsect .LT. 0.0) THEN
	
		max1tpsect=max1tpsect+2*pi
		
	END IF
	
	
	!For the second maximum:
	WRITE(20,*) 'For the second maximum:'
	tan_max=0.0
	h_max=gvm2(work_critp(locmax2))
	
	tan_tp=dgvm2(work_critp(locmax2-1))
	h_tp=gvm2(work_critp(locmax2-1))-tan_tp*work_critp(locmax2-1)
	WRITE(20,*) tan_tp, h_tp, tan_max, h_max
	CALL intersect(tan_tp, h_tp, tan_max, h_max, tpmax2sect)
	IF (tpmax2sect .LT. 0.0) THEN
	
		tpmax2sect=tpmax2sect+2*pi
		
	END IF
	WRITE(20,*) tan_tp, h_tp, tan_max, h_max, tpmax2sect
	IF (locmax2 .EQ. 8) THEN
	
		tan_tp=dgvm2(work_critp(1)+2*pi)
		h_tp=gvm2(work_critp(1)+2*pi)-tan_tp*(work_critp(1)+2*pi)
		WRITE(20,*) 'TP:', tan_tp, h_tp
	ELSE
	
		tan_tp=dgvm2(work_critp(locmax2+1))
		h_tp=gvm2(work_critp(locmax2+1))-tan_tp*work_critp(locmax2+1)
		WRITE(20,*) 'TP:', tan_tp, h_tp
	END IF
	WRITE(20,*) tan_tp, h_tp, tan_max, h_max
	CALL intersect(tan_max, h_max, tan_tp, h_tp, max2tpsect)
	IF (max2tpsect .LT. 0.0) THEN
	
		max2tpsect=max2tpsect+2*pi
		
	END IF
	WRITE(20,*) tan_tp, h_tp, tan_max, h_max, max2tpsect
	
	!Now write the vector with all necessary points for the approximation:
	
	DO jj=1, locmax1-1
	
		critp(jj)=work_critp(jj)
		fcritp(jj)=gvm2(critp(jj))
		
	END DO
	
	DO jj=locmax1+1, locmax2-1
	
		critp(jj)=work_critp(jj)
		fcritp(jj)=gvm2(critp(jj))
		
	END DO
	
	DO jj=locmax2+1, nex+ntp
	
		critp(jj)=work_critp(jj)
		fcritp(jj)=gvm2(critp(jj))
		
	END DO
	
	critp(locmax1)=tpmax1sect
	fcritp(locmax1)=gvm2(work_critp(locmax1))
	critp(nex+ntp+1)=max1tpsect
	fcritp(nex+ntp+1)=gvm2(work_critp(locmax1))
	critp(locmax2)=tpmax2sect
	fcritp(locmax2)=gvm2(work_critp(locmax2))
	critp(nex+ntp+2)=max2tpsect
	fcritp(nex+ntp+2)=gvm2(work_critp(locmax2))
	
	!Finish by sorting the vector in ascending value
	
	DO jj=1, nex+ntp+1
	
		DO kk=jj+1, nex+ntp+2
		
		  IF (critp(jj) .GT. critp(kk)) THEN
		  
			 wrk=critp(kk)
			 fwrk=fcritp(kk)
			 critp(kk)=critp(jj)
			 fcritp(kk)=fcritp(jj)
			 critp(jj)=wrk
			 fcritp(jj)=fwrk
			 
		  END IF
		  
		END DO
		
	END DO
	
	OPEN(UNIT=007, FILE='gvm2critpoints.txt', STATUS='unknown')
	
	DO jj=1, nex+ntp+2
	
		WRITE(007,*) critp(jj), fcritp(jj)
		
	END DO
	
	CLOSE(007)
	
	
	END SUBROUTINE critical_points
	


	!subroutines computing intersection of straight lines and their parameters:


	SUBROUTINE mhfromPQ(P, Q, m, h)
	IMPLICIT NONE
	
	DOUBLE PRECISION, INTENT(IN), DIMENSION(2)						:: P, Q
	DOUBLE PRECISION, INTENT(OUT)							:: m, h
	
	m=(P(2)-Q(2))/(P(1)-Q(1))
	h=(Q(1)*P(2)-Q(2)*P(1))/(Q(1)-P(1))
	
	END SUBROUTINE mhfromPQ
	
	
	
	SUBROUTINE mhfromxxyy(x1, y1, x2, y2, m, h)
	IMPLICIT NONE
	
	DOUBLE PRECISION								:: x1, y1, x2, y2
	DOUBLE PRECISION								:: m, h
	
	m=(y1-y2)/(x1-x2)
	h=y1-x1*m!(x2*y1-y2*x1)/(x2-x1)
	
	END SUBROUTINE mhfromxxyy
	
	
	
	SUBROUTINE intersect(m1, h1, m2, h2, xsect)
	IMPLICIT NONE
	
	DOUBLE PRECISION								:: m1, m2, h1, h2
	DOUBLE PRECISION								:: xsect

	IF (ABS(h2-h1) .LT. 1.0e-5) THEN
	
		IF (h2-h1 .GT. 0.0) THEN
		
			xsect=1.0e-5
		  
		ELSE
		
			xsect=-1.0e-5
		  
		END IF
		
	ELSE
	
		xsect=-(h2-h1)/(m2-m1)!(h2-h1)/(m1-m2)
	
	END IF
	
	IF (ABS(m2-m1) .LT. 1.0e-5) THEN
	
		IF (m2-m1 .LT. 0.0) THEN
		
			xsect=m1-1.0e-4
		  
		ELSE
		
			xsect=m2+1.0e-4
		  
		END IF
		
	END IF
	
	
	END SUBROUTINE intersect


	!Generalized von Mises of order 2 approximation function


	FUNCTION gvm2approx(x)
	IMPLICIT NONE
	
	DOUBLE PRECISION				:: gvm2approx
	DOUBLE PRECISION				:: x, dummy
	!IMPORTANT NOTE: the vector with the pertinent points has to be specified in the sub-program in which gvm2approx is called
	DOUBLE PRECISION, DIMENSION(10)			:: points, fpoints
	DOUBLE PRECISION				:: x1, x2, y1, y2, m, h
	INTEGER						:: npoints, jj
	
	npoints=SIZE(points)
	
	OPEN(UNIT=007, FILE='gvm2critpoints.txt', STATUS='unknown')
	
	DO jj=1,10
	
		READ(007,*) points(jj), fpoints(jj)
		
	END DO
	
	CLOSE(007)
	
	IF (x .GE. 0.0 .AND. x .LT. points(1)) THEN
	
		x1=points(10)-2*pi
		y1=fpoints(10)
		x2=points(1)
		y2=fpoints(1)
		CALL mhfromxxyy(x1, y1, x2, y2, m, h)
		
	ELSE IF (x .GE. points(1) .AND. x .LT. points(2)) THEN
	
		x1=points(1)
		y1=fpoints(1)
		x2=points(2)
		y2=fpoints(2)
		CALL mhfromxxyy(x1, y1, x2, y2, m, h)
		
	ELSE IF (x .GE. points(2) .AND. x .LT. points(3)) THEN
	
		x1=points(2)
		y1=fpoints(2)
		x2=points(3)
		y2=fpoints(3)
		CALL mhfromxxyy(x1, y1, x2, y2, m, h)
		
	ELSE IF (x .GE. points(3) .AND. x .LT. points(4)) THEN
	
		x1=points(3)
		y1=fpoints(3)
		x2=points(4)
		y2=fpoints(4)
		CALL mhfromxxyy(x1, y1, x2, y2, m, h)
		
	ELSE IF (x .GE. points(4) .AND. x .LT. points(5)) THEN
	
		x1=points(4)
		y1=fpoints(4)
		x2=points(5)
		y2=fpoints(5)
		CALL mhfromxxyy(x1, y1, x2, y2, m, h)
		
	ELSE IF (x .GE. points(5) .AND. x .LT. points(6)) THEN
	
		x1=points(5)
		y1=fpoints(5)
		x2=points(6)
		y2=fpoints(6)
		CALL mhfromxxyy(x1, y1, x2, y2, m, h)
		
	ELSE IF (x .GE. points(6) .AND. x .LT. points(7)) THEN
	
		x1=points(6)
		y1=fpoints(6)
		x2=points(7)
		y2=fpoints(7)
		CALL mhfromxxyy(x1, y1, x2, y2, m, h)
		
	ELSE IF (x .GE. points(7) .AND. x .LT. points(8)) THEN
	
		x1=points(7)
		y1=fpoints(7)
		x2=points(8)
		y2=fpoints(8)
		CALL mhfromxxyy(x1, y1, x2, y2, m, h)
		
	ELSE IF (x .GE. points(8) .AND. x .LT. points(9)) THEN
	
		x1=points(8)
		y1=fpoints(8)
		x2=points(9)
		y2=fpoints(9)
		CALL mhfromxxyy(x1, y1, x2, y2, m, h)
		
	ELSE IF (x .GE. points(9) .AND. x .LT. points(10)) THEN
	
		x1=points(9)
		y1=fpoints(9)
		x2=points(10)
		y2=fpoints(10)
		CALL mhfromxxyy(x1, y1, x2, y2, m, h)
		
	ELSE IF (x .GE. points(10) .AND. x .LE. 2*pi) THEN
	
		x1=points(10)
		y1=fpoints(10)
		x2=points(1)+2*pi
		y2=fpoints(1)
		CALL mhfromxxyy(x1, y1, x2, y2, m, h)

	ELSE
	
		WRITE(*,*) 'Function argument is not in [0,2*Pi)'
		
	END IF
	
	gvm2approx=m*x+h
	!WRITE(*,*) x, gvm2approx
	
	END FUNCTION


	!improved acceptance-rejection algorithm and GvM2 approximation

	
	SUBROUTINE iacc_rej(nacc, pdf)
	USE numint
	IMPLICIT NONE
	DOUBLE PRECISION					:: dummyu, dummyv, u, v, pdf, m, crit, simqual, dummyw, tin, tout, twopi=2.0d0*pi, ntot
	DOUBLE PRECISION, DIMENSION(10)				:: points, fpoints
	INTEGER							:: jj, kk, nacc, nrej, nsim
	DOUBLE PRECISION, DIMENSION(nacc)			:: simval, fsimval
	EXTERNAL pdf

	CALL CPU_time(tin)	
	OPEN(UNIT=007, FILE='gvm2critpoints.txt', STATUS='unknown')
	
	DO jj=1, 10
	
	   READ(007,*) points(jj), fpoints(jj)
	   
	END DO
	
	CLOSE(007)
	
	OPEN(UNIT=2800, FILE='statistics_iaccrej.txt', STATUS='unknown')
	OPEN(UNIT=2900, FILE='data_iaccrej.txt', STATUS='unknown')
	
	m=1.0d0

	!DO jj=1, 9

		!m=0.5d0*(fpoints(jj)+fpoints(jj+1))*(points(jj+1)-points(jj))

	!END DO
	nsim=0
	nrej=0
	
	DO WHILE(nsim-nrej .LT. nacc)
	
	   nsim=nsim+1
	   CALL random_approxgvm2(dummyu,ntot)
		!WRITE(*,*) 'Total surface area of the approximation:', ntot
	   CALL random_number(dummyv)
	   
	   u=dummyu
	   v=dummyv*m*gvm2approx(u)/ntot
	
	   crit=pdf(u)/m/gvm2approx(u)

		WRITE(20,*) 'u=', u, 'gvm2=', pdf(u), 'approx=', gvm2approx(u), 'V=', dummyv
	   
	   IF (dummyv .LE. crit) THEN
	   
	      simval(nsim-nrej)=u
	      fsimval(nsim-nrej)=v
	      
	   ELSE
	   
	      nrej=nrej+1
	      
	   END IF
	   
	END DO
	
	!CALL sort(simval)
	
	DO jj=1, nacc
	
	   WRITE(2900,*) simval(jj), fsimval(jj)
	   
	END DO
	
	CALL CPU_time(tout)
	WRITE(2800,*) 'Statistics for the improved acceptance-rejection method:'
	WRITE(2800,*) 'Number of simulated random variables:', 2*nsim
	WRITE(2800,*) 'Number of rejections:', 2*nrej
	WRITE(2800,*) 'Number of acceptances:', 2*nacc
	simqual=nacc*1.0/(nsim*1.0)
	WRITE(2800,*) 'Percentage of accepted rv per simulated rv:', simqual
	WRITE(2800,*) 'The method took:', tout-tin, 'seconds'	

	CLOSE(2900)
	CLOSE(2800)
		
	
	
	
	END SUBROUTINE iacc_rej
	
	
	SUBROUTINE random_approxgvm2(out,ntot)
	IMPLICIT NONE
	DOUBLE PRECISION, DIMENSION(10)				:: points, fpoints
	DOUBLE PRECISION, DIMENSION(11)				:: weight, pente, hauteur, normale
	INTEGER							:: jj, z
	DOUBLE PRECISION					:: out, dummy, x, x1, x2, y1, y2, m, h, dummyw, wtot, allpoints(12), ntot, root
	
	OPEN(UNIT=007, FILE='gvm2critpoints.txt', STATUS='unknown')
	
	allpoints(1)=0.0d0
	allpoints(12)=2*pi
	DO jj=1, 10
	
	   READ(007,*) points(jj), fpoints(jj)
	   allpoints(jj+1)=points(jj)

	END DO
	
	CLOSE(007)


	CALL mhfromxxyy(points(10)-2*pi, fpoints(10), points(1), fpoints(1), pente(1), hauteur(1))

	DO jj=1,9

		CALL mhfromxxyy(points(jj), fpoints(jj), points(jj+1), fpoints(jj+1), pente(jj+1), hauteur(jj+1))

	END DO

	CALL mhfromxxyy(points(10), fpoints(10), points(1)+2*pi, fpoints(1), pente(11), hauteur(11))

	ntot=0.0d0
	DO jj=1,11

		normale(jj)=0.5*pente(jj)*(allpoints(jj+1)**2-allpoints(jj)**2)+hauteur(jj)*(allpoints(jj+1)-allpoints(jj))
		ntot=ntot+normale(jj)

	END DO

	wtot=0.0d0
	DO jj=1,11

		weight(jj)=normale(jj)/ntot
		wtot=wtot+weight(jj)

	END DO

	CALL random_number(dummyw)

	wtot=0.0
	DO jj=1,11

		wtot=wtot+weight(jj)
		IF (dummyw .LT. wtot) THEN

			z=jj
			EXIT

		END IF

	END DO

	CALL random_number(dummy)

	IF (Z .EQ. 1) THEN

		IF (ABS(pente(1)) .LE. 1.0d-9) THEN

			out=dummy*(allpoints(2)-allpoints(1))+allpoints(1)

		ELSE

			root=SQRT(hauteur(1)**2+2*pente(1)*normale(1)*dummy)
			out=(-hauteur(1)+root)/pente(1)

			IF (out .LT. 0.0d0) THEN

				out=(-hauteur(1)-root)/pente(1)

			END IF

		END IF

	ELSE IF (Z .EQ. 11) THEN

		IF (ABS(pente(11)) .LE. 1.0d-9) THEN

			out=dummy*(allpoints(12)-allpoints(11))+allpoints(11)

		ELSE

			root=SQRT(hauteur(11)**2+2*pente(11)*normale(11)*dummy+pente(11)**2*allpoints(11)**2+2*pente(11)*hauteur(11)*allpoints(11))
			out=(-hauteur(11)+root)/pente(11)

			IF (out .LT. 0.0d0) THEN

				out=(-hauteur(11)-root)/pente(11)

			END IF

		END IF

	ELSE 

		IF (ABS(pente(Z)) .LE. 1.0d-9) THEN

			out=dummy*(allpoints(Z+1)-allpoints(Z))+allpoints(Z)

		ELSE

			root=SQRT(hauteur(Z)**2+2*pente(Z)*normale(Z)*dummy+pente(Z)**2*allpoints(Z)**2+2*pente(Z)*hauteur(Z)*allpoints(Z))
			out=(-hauteur(Z)+root)/pente(Z)

			IF (out .LT. 0.0d0) THEN

				out=(-hauteur(Z)-root)/pente(Z)

			END IF

		END IF

	END IF
		

	WRITE(20,*) dummy, z, out

	END SUBROUTINE random_approxgvm2
	


	
	!subroutine for sorting arrays


	SUBROUTINE sort(vector)
	IMPLICIT NONE
	
	DOUBLE PRECISION, DIMENSION(:)						:: vector
	DOUBLE PRECISION, DIMENSION(:),ALLOCATABLE			:: workspace
	DOUBLE PRECISION									 :: wrk
	INTEGER								  :: dimvec, jj, kk
	
	dimvec=SIZE(vector)
	ALLOCATE(workspace(dimvec))
	
	DO jj=1, dimvec-1
	
		DO kk=jj+1, dimvec
		
		  IF (vector(jj) .GT. vector(kk)) THEN
		  
			 wrk=vector(kk)
			 vector(kk)=vector(jj)
			 vector(jj)=wrk
			 
		  END IF
		  
		END DO
		
	END DO
	
	END SUBROUTINE sort
	



END PROGRAM sim_gvm2


	DOUBLE PRECISION FUNCTION rawGvM2(x)
	USE constants
	IMPLICIT NONE
	
	DOUBLE PRECISION				:: x
	
	!OPEN(UNIT=1000, FILE='GvMparam.txt', STATUS='old')
	!READ(1000,*) k1, k2, m1, m2
	!CLOSE(1000)
	
	IF (x .lt. 0.0 .or. x .gt. 2*pi) THEN
	
		OPEN(UNIT=9, FILE="ERRORLOG.txt", STATUS="unknown")
		WRITE(9,*) 'ERROR at rawGvM2: x out of bound', x
		CLOSE(9)
		
		STOP
		
	ELSE
	
		rawGvM2=EXP(k1*COS(x-m1)+k2*COS(2*(x-m2)))
		
	ENDIF
	
	END FUNCTION rawGvM2



	FUNCTION gvm2(x)
	USE constants
	!USE numint
	IMPLICIT NONE

	DOUBLE PRECISION, INTENT(in)				:: x
	DOUBLE PRECISION					:: gvm2, norm, rawgvm2, twopi
	!DOUBLE PRECISION					:: kappa1, kappa2, mu1, mu2, delta
	external rawgvm2

	!OPEN(UNIT=1000, FILE="GvMparam.txt", STATUS="old")
	!READ(1000,*) kappa1, kappa2, mu1, mu2, delta, norm
	!CLOSE(1000)
	!twopi=2.0d0*pi
	!CALL trapint(rawgvm2, 0.0d0, twopi, 1d-6, norm)
	!WRITE(*,*) 'Gvm2 parameters:', k1, k2, m1, m2
	gvm2=EXP(k1*COS(x-m1)+k2*COS(2*(x-m2)))!/norm

	END FUNCTION gvm2



	FUNCTION dgvm2(x)
	USE constants
	!USE numint
	IMPLICIT NONE

	DOUBLE PRECISION, INTENT(in)				:: x
	DOUBLE PRECISION					:: dgvm2, exponential, differential, norm, rawgvm2, twopi=2.0d0*pi
	!DOUBLE PRECISION					:: kappa1, kappa2, mu1, mu2, delta
	EXTERNAL rawgvm2

	!OPEN(UNIT=1000, FILE="GvMparam.txt", STATUS="old")
	!READ(1000,*) kappa1, kappa2, mu1, mu2, delta, norm
	!CLOSE(1000)
	!WRITE(*,*) 'dGvm2 parameters:', k1, k2, m1, m2
	exponential=EXP(k1*COS(x-m1)+k2*COS(2*(x-m2)))
	differential=-k1*SIN(x-m1)-2.0*k2*SIN(2*(x-m2))

	!CALL trapint(rawgvm2,0.0d0,twopi,1d-6,norm)

	dgvm2=exponential*differential!/norm

	END FUNCTION dgvm2



	FUNCTION ddgvm2(x)
	USE constants
	!USE numint
	IMPLICIT NONE

	DOUBLE PRECISION, INTENT(in)				:: x
	DOUBLE PRECISION					:: ddgvm2, exponential, diffsquared, doublediff, norm, rawgvm2, twopi=2.0d0*pi
	!DOUBLE PRECISION					:: kappa1, kappa2, mu1, mu2, delta
	EXTERNAL rawgvm2

	!OPEN(UNIT=1000, FILE="GvMparam.txt", STATUS="old")
	!READ(1000,*) kappa1, kappa2, mu1, mu2, delta, norm
	!CLOSE(1000)
	!WRITE(*,*) 'ddGvm2 parameters:', k1, k2, m1, m2
	exponential=EXP(k1*COS(x-m1)+k2*COS(2*(x-m2)))
	diffsquared=(-k1*SIN(x-m1)-2.0*k2*SIN(2*(x-m2)))**2
	doublediff=-k1*COS(x-m1)-4.0*k2*COS(2*(x-m2))

	!CALL trapint(rawgvm2,0.0d0,twopi,1d-6,norm)

	ddgvm2=exponential*(diffsquared+doublediff)!/norm

	END FUNCTION ddgvm2

